""" Navigation display for the QTGL gui."""
from PyQt6.QtCore import qCritical, QTimer
from PyQt6.QtOpenGLWidgets import QOpenGLWidget as QGLWidget
import OpenGL.GL as gl
from math import sin, cos, radians
import numpy as np
from ctypes import c_float, c_int, Structure

import bluesky as bs
from .glhelpers import BlueSkyProgram, RenderObject, UniformBuffer

VERTEX_IS_LATLON, VERTEX_IS_METERS, VERTEX_IS_SCREEN, VERTEX_IS_GLXY = list(range(4))
ATTRIB_VERTEX, ATTRIB_TEXCOORDS, ATTRIB_LAT, ATTRIB_LON, ATTRIB_ORIENTATION, ATTRIB_COLOR, ATTRIB_TEXDEPTH = list(range(7))

white                = (255, 255, 255)
black                = (0,   0,   0)
yellow               = (255, 255, 0)
green                = (0,   255, 0)
lightblue2           = (148, 178, 235)
lightblue3           = (220, 250, 255)


class ndUBO(UniformBuffer):
    class Data(Structure):
        _fields_ = [("ownhdg", c_float), ("ownlat", c_float), ("ownlon", c_float),
        ("zoom", c_float), ("ownid", c_int), ("vertex_modifiers", c_int)]

    data = Data(0.0, 0.0, 0.0, 4.0, 3)

    def __init__(self):
        super().__init__(self.data)

    def set_zoom(self, zoom):
        self.data.zoom   = zoom

    def set_owndata(self, acid, lat, lon, hdg):
        self.data.ownid  = acid
        self.data.ownlat = lat
        self.data.ownlon = lon
        self.data.ownhdg = hdg

    def set_vertex_modifiers(self, scale_type, rotate_ownhdg):
        self.data.vertex_modifiers = scale_type + (10 if rotate_ownhdg else 0)
        self.update()


class ND(QGLWidget):
    def __init__(self, parent=None, shareWidget=None):
        super().__init__(parent=parent, shareWidget=shareWidget)

        self.shareWidget = shareWidget
        self.ac_id = ''
        self.n_aircraft = None
        self.initialized = False
        self.invalid_count = 0
        # Set size
        self.viewport = (0, 0, 400, 400)
        self.resize(400, 400)

        # Connect to incoming stream data
        bs.net.stream_received.connect(self.on_simstream_received)

    def on_simstream_received(self, streamname, data, sender_id):
        if streamname == b'ACDATA':
            if self.ac_id in data['id']:
                idx = data['id'].index(self.ac_id.upper())
                lat = data['lat'][idx]
                lon = data['lon'][idx]
                trk = data['trk'][idx]
                tas = data['tas'][idx]
                self.n_aircraft = len(data['lat'])
                self.globaldata.set_owndata(idx, lat, lon, trk)

    def setAircraftID(self, ac_id):
        self.ac_id = ac_id
        self.setWindowTitle(ac_id)

    def create_objects(self):
        if not (self.isValid() and self.shareWidget.initialized):
            self.invalid_count += 1
            print('ND: Context not valid in create_objects, or shareWidget not yet initialized')
            QTimer.singleShot(100, self.create_objects)
            return

        # Make the nd widget context current, necessary when create_objects is not called from initializeGL
        self.makeCurrent()

        # Use the same font as the radarwidget
        self.font = self.shareWidget.font.copy()
        self.font.init_shader(self.text_shader)

        edge = np.zeros(120, dtype=np.float32)
        edge[0:120:2] = 1.4 * np.sin(np.radians(np.arange(-60, 60, 2)))
        edge[1:120:2] = 1.4 * np.cos(np.radians(np.arange(-60, 60, 2)))
        self.edge = RenderObject(gl.GL_LINE_STRIP, vertex=edge, color=white)

        arcs = []
        for i in range(1, 4):
            for angle in range(-60, 60, max(2, 6 - 2 * i)):
                arcs.append(float(i) * 0.35 * sin(radians(angle)))
                arcs.append(float(i) * 0.35 * cos(radians(angle)))
                if i == 4:
                    arcs.append(float(i) * 0.35 * sin(radians(angle + 2)))
                    arcs.append(float(i) * 0.35 * cos(radians(angle + 2)))
        arcs = np.array(arcs, dtype=np.float32)
        self.arcs = RenderObject(gl.GL_LINES, vertex=arcs, color=white)

        mask = []
        for angle in range(-60, 60, 2):
            mask.append(1.4 * sin(radians(angle)))
            mask.append(10.0)
            mask.append(1.4 * sin(radians(angle)))
            mask.append(1.4 * cos(radians(angle)))
        mask = np.array(mask, dtype=np.float32)
        self.mask = RenderObject(gl.GL_TRIANGLE_STRIP, vertex=mask, color=black)

        ticks = np.zeros(288, dtype=np.float32)
        for i in range(72):
            ticktop = 1.46 if i % 6 == 0 else (1.44 if i % 2 == 0 else 1.42)
            ticks[4*i  :4*i+2] = (1.4 * sin(radians(i * 5)), 1.4 * cos(radians(i * 5)))
            ticks[4*i+2:4*i+4] = (ticktop * sin(radians(i * 5)), ticktop * cos(radians(i * 5)))
        self.ticks = RenderObject(gl.GL_LINES, vertex=ticks, color=white)

        ticklbls = np.zeros(24 * 36, dtype=np.float32)
        texcoords = np.zeros(36 * 36, dtype=np.float32)

        for i in range(36):
            if i % 3 == 0:
                w, h, y = 0.045, 0.09, 1.48
            else:
                w, h, y = 0.035, 0.07, 1.46
            tmp = [(-w, h+y), (-w, y), (0.0, h+y), (0.0, h+y), (-w, y), (0.0, y),
                   (0.0, h+y), (0.0, y), (w, h+y), (w, h+y), (0.0, y), (w, y)]

            # numerics start at ASCII 48
            c1 = i / 10 + 48
            c2 = i % 10 + 48
            texcoords[36*i:36*i+18]    = [0, 0, c1, 0, 1, c1, 1, 0, c1, 1, 0, c1, 0, 1, c1, 1, 1, c1]
            texcoords[36*i+18:36*i+36] = [0, 0, c2, 0, 1, c2, 1, 0, c2, 1, 0, c2, 0, 1, c2, 1, 1, c2]
            angle = radians(10 * (36 - i))
            rot = np.array([[cos(angle), -sin(angle)], [sin(angle), cos(angle)]])
            for j in range(12):
                ticklbls[24*i+2*j:24*i+2*j+2] = rot.dot(tmp[j])

        self.ticklbls = RenderObject(gl.GL_TRIANGLES, vertex=ticklbls, color=white, texcoords=texcoords)

        vown = np.array([0.0, 0.0, 0.0, -0.12, 0.065, -0.03, -0.065, -0.03, 0.022, -0.1, -0.022, -0.1], dtype=np.float32)
        self.ownship = RenderObject(gl.GL_LINES, vertex=vown, color=yellow)

        self.spdlabel_text = self.font.prepare_text_string('GS    TAS', 0.05, white, (-0.98, 1.6))
        self.spdlabel_val  = self.font.prepare_text_string('  000    000', 0.05, green, (-0.97, 1.6))

        self.waypoints = RenderObject.copy(self.shareWidget.waypoints)
        self.wptlabels = RenderObject.copy(self.shareWidget.wptlabels)
        self.airports  = RenderObject.copy(self.shareWidget.airports)
        self.aptlabels = RenderObject.copy(self.shareWidget.aptlabels)
        self.protectedzone = RenderObject.copy(self.shareWidget.protectedzone)
        self.ac_symbol = RenderObject.copy(self.shareWidget.ac_symbol)
        self.aclabels = RenderObject.copy(self.shareWidget.aclabels)

        # Unbind VAO, VBO
        RenderObject.unbind_all()

        # Done initializing
        self.initialized = True

    def initializeGL(self):
        """Initialize OpenGL, VBOs, upload data on the GPU, etc."""

        # First check for supported GL version
        gl_version = float(gl.glGetString(gl.GL_VERSION)[:3])
        if gl_version < 3.3:
            return

        # background color
        gl.glClearColor(0, 0, 0, 0)
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        self.globaldata = ndUBO()

        try:
            # Compile shaders and link color shader program
            self.color_shader = BlueSkyProgram('graphics/shaders/nd-normal.vert', 'graphics/shaders/nd-color.frag')
            self.color_shader.bind_uniform_buffer('global_data', self.globaldata)

            # Compile shaders and link text shader program
            self.text_shader = BlueSkyProgram('graphics/shaders/nd-text.vert', 'graphics/shaders/nd-text.frag')
            self.text_shader.bind_uniform_buffer('global_data', self.globaldata)

        except RuntimeError as e:
            qCritical('Error compiling shaders in radarwidget: ' + e.args[0])
            return

        # Set initial zoom
        self.globaldata.set_zoom(4.0)

        self.create_objects()

    def resizeGL(self, width, height):
        # paint within the largest possible rectangular area in the window
        w = h = min(width, height)
        x = max(0, (width - w) // 2)
        y = max(0, (height - h) // 2)
        self.viewport = (x, y, w, h)

    def paintGL(self):
        """Paint the scene."""
        # pass if the framebuffer isn't complete yet or if not initialized
        if not (gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER) == gl.GL_FRAMEBUFFER_COMPLETE and self.initialized and self.isVisible()):
            return

        # Set the viewport and clear the framebuffer
        gl.glViewport(*self.viewport)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        # Select the non-textured shader
        self.color_shader.use()

        self.globaldata.set_vertex_modifiers(VERTEX_IS_GLXY, False)
        self.arcs.draw()

        self.globaldata.set_vertex_modifiers(VERTEX_IS_METERS, False)
        self.protectedzone.draw(n_instances=self.n_aircraft)
        self.globaldata.set_vertex_modifiers(VERTEX_IS_SCREEN, True)
        self.ac_symbol.draw(n_instances=self.n_aircraft)

        self.globaldata.set_vertex_modifiers(VERTEX_IS_SCREEN, False)
        self.waypoints.bind()
        gl.glVertexAttrib4Nub(ATTRIB_COLOR, *(lightblue2 + (255,)))
        self.waypoints.draw()
        self.airports.bind()
        gl.glVertexAttrib4Nub(ATTRIB_COLOR, *(lightblue2 + (255,)))
        self.airports.draw()

        self.text_shader.use()
        self.font.use()
        self.font.set_char_size(self.wptlabels.char_size)
        self.font.set_block_size(self.wptlabels.block_size)
        self.wptlabels.bind()
        gl.glVertexAttrib4Nub(ATTRIB_COLOR, *(lightblue3 + (255,)))
        self.wptlabels.draw(n_instances=self.waypoints.n_instances)

        self.font.set_char_size(self.aptlabels.char_size)
        self.font.set_block_size(self.aptlabels.block_size)
        self.aptlabels.bind()
        gl.glVertexAttrib4Nub(ATTRIB_COLOR, *(lightblue3 + (255,)))
        self.aptlabels.draw(n_instances=self.airports.n_instances)

        self.font.set_char_size(self.aclabels.char_size)
        self.font.set_block_size(self.aclabels.block_size)
        self.aclabels.draw(n_instances=self.n_aircraft)

        self.color_shader.use()
        self.globaldata.set_vertex_modifiers(VERTEX_IS_GLXY, False)
        self.ownship.draw()
        self.mask.draw()
        self.edge.draw()

        self.globaldata.set_vertex_modifiers(VERTEX_IS_GLXY, True)
        self.ticks.draw()

        # Select the text shader
        self.text_shader.use()
        self.font.use()
        self.font.set_block_size((0, 0))
        self.ticklbls.draw()

        self.globaldata.set_vertex_modifiers(VERTEX_IS_GLXY, False)
        self.spdlabel_text.draw()
        self.spdlabel_val.draw()

        # Unbind everything
        RenderObject.unbind_all()
        gl.glUseProgram(0)



""" Classes that derive from TrafficArrays (like Traffic) get automated create,
    delete, and reset functionality for all registered child arrays."""
# -*- coding: utf-8 -*-
from collections.abc import Collection
import numpy as np

defaults = {"float": 0.0, "int": 0, "uint":0, "bool": False, "S": "", "str": ""}


class RegisterElementParameters:
    """ Class to use in 'with'-syntax. This class automatically
        calls for the _init_trafarrays function of the
        DynamicArray, with all parameters defined in 'with'."""

    def __init__(self, parent):
        self._parent = parent
        self.keys0 = set(parent.__dict__.keys())

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, tb):
        self._parent._init_trafarrays(set(self._parent.__dict__.keys()) - self.keys0)


class TrafficArrays:
    """ Parent class to use separate arrays and lists to allow
        vectorizing but still maintain and object like benefits
        for creation and deletion of an element for all parameters"""

    # The TrafficArrays class keeps track of all of the constructed
    # TrafficArray objects
    root = None
    ntraf = 0

    @staticmethod
    def setroot(obj):
        ''' This function is used to set the root of the tree of TrafficArray
            objects (which is the traffic object.)'''
        TrafficArrays.root = obj

    def __init__(self):
        super().__init__()
        self._parent   = TrafficArrays.root
        if self._parent:
            self._parent._children.append(self)
        self._children = []
        self._ArrVars  = []
        self._LstVars  = []

    def reparent(self, newparent):
        ''' Give TrafficArrays object a new parent. '''
        # Remove myself from the parent list of children, and add to new parent
        self._parent._children.pop(self._parent._children.index(self))
        newparent._children.append(self)
        self._parent = newparent

    def settrafarrays(self):
        ''' Convenience function for with-style traffic array registration. '''
        return RegisterElementParameters(self)

    def _init_trafarrays(self, keys):
        for key in keys:
            if isinstance(self.__dict__[key], list):
                self._LstVars.append(key)
            elif isinstance(self.__dict__[key], np.ndarray):
                self._ArrVars.append(key)
            elif isinstance(self.__dict__[key], TrafficArrays):
                self.__dict__[key].reparent(self)

        # In plugins and replaceable classes it could be that their instance
        # is created when the simulation is already running, and traffic is
        # present. Size traffic arrays accordingly here
        if TrafficArrays.root.ntraf:
            self.create(TrafficArrays.root.ntraf)

    def create(self, n=1):
        ''' Append n elements (aircraft) to all lists and arrays. '''

        for v in self._LstVars:  # Lists (mostly used for strings)
            lst: list = self.__dict__[v]
            vartype = type(lst[0]).__name__ if lst else 'str'
            lst.extend([defaults.get(vartype)] * n)

        for v in self._ArrVars:  # Numpy array
            # Get type without byte length
            vartype = ''.join(c for c in str(self.__dict__[v].dtype) if c.isalpha())
            self.__dict__[v] = np.append(self.__dict__[v], [defaults.get(vartype, 0)] * n)

    def istrafarray(self, name):
        ''' Returns true if parameter 'name' is a traffic array. '''
        return name in self._LstVars or name in self._ArrVars

    def create_children(self, n=1):
        ''' Call create (aircraft create) on all children. '''
        for child in self._children:
            child.create(n)
            child.create_children(n)

    def delete(self, idx):
        ''' Aircraft delete. '''
        # Remove element (aircraft) idx from all lists and arrays
        for child in self._children:
            child.delete(idx)

        for v in self._ArrVars:
            self.__dict__[v] = np.delete(self.__dict__[v], idx)

        if self._LstVars:
            if isinstance(idx, Collection):
                for i in reversed(idx):
                    for v in self._LstVars:
                        del self.__dict__[v][i]
            else:
                for v in self._LstVars:
                    del self.__dict__[v][idx]

    def reset(self):
        ''' Delete all elements from arrays and start at 0 aircraft. '''
        for child in self._children:
            child.reset()

        for v in self._ArrVars:
            self.__dict__[v] = np.array([], dtype=self.__dict__[v].dtype)

        for v in self._LstVars:
            self.__dict__[v] = []



import random
from math import degrees,radians,cos,sin,atan2,sqrt


from bluesky import stack,traf,sim,tools,navdb
from bluesky.tools.position import txt2pos
from bluesky.tools.geo import kwikqdrdist,kwikpos,latlondist,qdrdist
from bluesky.tools.misc import degto180,txt2alt,txt2spd
from bluesky.tools.aero import nm,ft

# Default values
swcircle = False
ctrlat = 52.6   # [deg]
ctrlon = 5.4    # [deg]
radius = 230.   # [nm]

def setcircle(ictrlat,ictrlon,iradius):
    global ctrlat,ctrlon,radius,swcircle
    swcircle = True
    ctrlat, ctrlon, radius = ictrlat,ictrlon,iradius
    return


class Source():
    # Define a source
    def __init__(self,name,cmd,cmdargs):
        self.name    = name.upper()
        self.tupdate = sim.simt-999.
        self.flow    = 0
        self.incircle = True
        self.segdir  = None  # Segment direction in degrees
        self.hdg     = None # When runway is used as separate airport

        # Is location a circle segment?
        if swcircle and self.name[:4]=="SEGM":
            self.type = "seg"
            self.lat,self.lon,brg = getseg(self.name)
            pass

        else:
            success, posobj = txt2pos(name,ctrlat,ctrlon)
            if success:
                # for runway type, get heading as default optional argument for command line
                if posobj.type == "rwy":
                    aptname, rwyname = name.split('/RW')
                    rwyname = rwyname.lstrip('Y')
                    try:
                        self.hdg = navdb.rwythresholds[aptname][rwyname][2]
                    except:
                        self.hdg = None
                        pass
                else:
                    rwyhdg = None


                self.lat,self.lon = posobj.lat, posobj.lon
                self.type = posobj.type

                # If outside circle change lat,lon to edge of circle
                self.incircle = incircle(self.lat,self.lon)
                if not self.incircle:
                    self.type = "seg"
                    self.segdir,dist = qdrdist(ctrlat,ctrlon,posobj.lat,posobj.lon)
                    segname = "SEGM"+str(int(round(self.segdir)))
                    self.lat,self.lon,hdg = getseg(segname)

            else:
                print("ERROR: trafgen class Source called for "+name+". Position not found")
                self.lat,self.lon = 0.0,0.0


        # Aircraft types
        self.actypes = ["*"]

        # Runways
        self.runways    = [] # name
        self.rwylat     = []
        self.rwylon     = []
        self.rwyhdg     = []
        self.rwyline    = [] # number of aircraft waiting in line
        self.rwytotime  = [] # time of last takeoff

        # If source type is a runway, add it as the only runway
        if self.type=="rwy":
            self.runways = [rwyname]
            self.rwylat     = [self.lat]
            self.rwylon     = [self.lon]
            self.rwyhdg     = [int(rwyname.rstrip("LCR").lstrip("0"))*10.]
            self.rwyline    = [] # number of aircraft waiting in line
            self.rwytotime  = [-999] # time of last takeoff


        self.dtakeoff = 90. # sec take-off interval on one runway, default 90 sec

        # Destinations
        self.dest        = []
        self.destlat     = []
        self.destlon     = []
        self.desttype    = [] # "apt","wpt","seg"
        self.desthdg     = []  # if dest is a runway (for flights within FIR) or circle segment (source outside FIR)
        self.destactypes = []   # Types for this destinations ([]=use defaults for this source)

        #Names of drawing objects runways
        self.polys       = []    # Names of current runway polygons to remove when runways change

        # Limits on start values alt,spd,hdg
        self.startaltmin = None
        self.startaltmax = None
        self.startspdmin = None
        self.startspdmax = None
        self.starthdgmin = None
        self.starthdgmax = None

        return

    def setrunways(self,cmdargs):
        self.runways   = []
        self.rwylat    = []
        self.rwylon    = []
        self.rwyhdg    = []
        self.rwyline   = []
        self.rwytotime = []

        for runwayname in cmdargs:
            if runwayname[0] == "R":
                success, rwyposobj = txt2pos(self.name + "/" + runwayname, self.lat, self.lon)
            else:
                success,rwyposobj = txt2pos(self.name+"/RW"+runwayname,self.lat,self.lon)
            if success:
                self.runways.append(runwayname)
                self.rwylat.append(rwyposobj.lat)
                self.rwylon.append(rwyposobj.lon)
                rwyname = runwayname.upper().lstrip('RWY')
                #try:
                if True:
                    self.rwyhdg.append(navdb.rwythresholds[self.name][rwyname][2])
                    #if self.name=="EHAM":
                    #S    print("runway added with hdg:",self.rwyhdg[-1])
                #except:
                #    success = False
                self.rwyline.append(0)
                self.rwytotime.append(-999.)
            else:
                self.runways.append(runwayname)
                self.rwylat.append(self.lat)
                self.rwylon.append(self.lon)
                rwydigits = runwayname.lstrip("RWY").rstrip("LCR")
                self.rwyhdg.append(10. * int(rwydigits.rstrip("LCR").lstrip("0")))
                self.rwyline.append(0)
                self.rwytotime.append(-999.)

    def setalt(self,cmdargs):
        if len(cmdargs)==1:
            alt = txt2alt(cmdargs[0])
            self.startaltmin = alt
            self.startaltmax = alt
        elif len(cmdargs)>1:
            alt0,alt1 = txt2alt(cmdargs[0]),txt2alt(cmdargs[1])
            self.startaltmin = min(alt0,alt1)
            self.startaltmax = max(alt0,alt1)
        else:
            stack.stack("ECHO "+self.name+" ALT "+str(self.startaltmin)+" "+str(self.startaltmax))

    def setspd(self,cmdargs):
        if len(cmdargs)==1:
            spd = txt2spd(cmdargs[0])
            self.startspdmin = spd
            self.startapdmax = spd
        elif len(cmdargs)>1:
            spd0,spd1 = txt2spd(cmdargs[0]),txt2spd(cmdargs[1])
            self.startspdmin = min(spd0,spd1)
            self.startspdmax = max(spd0,spd1)
        else:
            stack.stack("ECHO "+self.name+" SPD "+str(self.startaltmin)+" "+str(self.startaltmax))

    def sethdg(self,cmdargs):
        if len(cmdargs)==1:
            hdg = float(cmdargs[0])
            self.starthdgmin = hdg
            self.starthdgmax = hdg
        elif len(cmdargs)>1:
            hdg0,hdg1 = float(cmdargs[0]),float(cmdargs[1])
            hdg0,hdg1 = min(hdg0,hdg1),max(hdg0,hdg1)
            if hdg1-hdg0>180.:
                hdg0,hdg1 = hdg1-360,hdg0
            self.starthdgmin = hdg0
            self.starthdgmax = hdg1
        else:
            stack.stack("ECHO ERROR "+self.name+" HDG "+str(self.starthdgmin)+" "+str(self.starthdgmax))

    def adddest(self,cmdargs):
        # Add destination with a given aircraft types
        destname = cmdargs[0]
        try:
            freq = int(cmdargs[1])
        except:
            freq = 1

        # Get a/c types frequency list, if given
        destactypes = []
        if len(cmdargs) > 2:  # also types are given for this destination
            for c in cmdargs[2:]:
                if c.count(":") > 0:
                    actype, typefreq = c.split(":")
                else:
                    actype = c
                    typefreq = "1"

                for f in range(int(typefreq)):
                    destactypes.append(actype)

        if not destname[:4]=="SEGM":
            success,posobj = txt2pos(destname,self.lat,self.lon)
            if success:
                if posobj.type == "rwy":
                    for i in range(freq):
                        self.dest.append(destname)
                        self.destlat.append(posobj.lat)
                        self.destlon.append(posobj.lon)
                        self.desthdg.append(None)
                        self.desttype.append(posobj.type)
                        self.destactypes.append(destactypes)

                else:
                    for i in range(freq):
                        aptname = destname
                        self.dest.append(destname)
                        self.destlat.append(posobj.lat)
                        self.destlon.append(posobj.lon)
                        self.desthdg.append(0)
                        self.desttype.append(posobj.type)
                        self.destactypes.append(destactypes)
            else:
                # Add random destination
                for i in range(freq):
                    name = "SEGM" + str(int(random.random() * 360.))
                    lat, lon, hdg = getseg(name)
                    self.dest.append(name)
                    self.destlat.append(lat)
                    self.destlon.append(lon)
                    self.desthdg.append(hdg)
                    self.desttype.append("seg")
                    self.destactypes.append(destactypes)

        else:
            for i in range(freq):
                # Segment as destination, bearing from center = heading
                lat,lon,hdg = getseg(destname)
                self.dest.append(destname)
                self.destlat.append(lat)
                self.destlon.append(lon)
                self.desthdg.append(hdg)
                self.desttype.append("seg")
                self.destactypes.append(destactypes)

        return True

    def setflow(self,flowtxt):
        try:
            self.flow = float(flowtxt) #in a/c per hour, also starts flow as it by default zero
        except:
            return False
        return True


    def addactypes(self,actypelist):
        self.actypes = self.actypes + makefreqlist(actypelist)

    def update(self,gain):
        # Time step update of source

        # Get time step
        dt = sim.simt - self.tupdate
        self.tupdate = sim.simt

        # Time for a new aircraft?
        if dt>0.0:

            # Calculate probability of a geberate occurring with flow
            chances = 1.0-gain*self.flow*dt/3600. #flow is in a/c per hour=360 seconds
            if random.random() >= chances:

                # Runways defined? => use runway lines (queues)
                if len(self.runways)>0:
                    # We do not yet need to create an aircraft
                    gennow = False

                    # Find shortest line and put it in
                    isel = random.randint(0,len(self.runways)-1)
                    self.rwyline[isel] = self.rwyline[isel] + 1
                else:
                    # Yes we do need to generate one now
                    gennow = True
            else:
                gennow = False

            # Check for generating aircraft
            # First check runways for a/c already in line:
            txt = ""
            for i in range(len(self.runways)):
                # Runway vacated and there is one waiting?
                if sim.simt-self.rwytotime[i]>self.dtakeoff and self.rwyline[i]>0:
                    #if self.name == "EHAM":
                    #   print(sim.simt, self.runways[i], self.rwytotime[i])
                    self.rwytotime[i] = sim.simt
                    self.rwyline[i]   = self.rwyline[i]-1

                    # Choose and aicraft type, check for distance
                    if len(self.dest)>0:
                        idest = int(random.random() * len(self.dest))
                    else:
                        idest = -1

                    if idest>=0:
                        acid = randacname(self.name,self.dest[idest])

                        if self.desttype[idest]=="seg" or self.dest[idest][:4]=="SEGM":
                            lat,lon,hdg = getseg(self.dest[idest])
                        else:
                            success,posobj = txt2pos(self.dest[idest],ctrlat,ctrlon)
                            lat,lon = posobj.lat,posobj.lon
                        distroute = latlondist(self.lat,self.lon,lat,lon)/nm

                    else:
                        acid = randacname(self.name, self.name)

                    if self.destactypes[idest] == []:
                        actype = random.choice(self.actypes)
                        actype = checkactype(actype, distroute, self.actypes)
                    else:
                        actype = random.choice(self.destactypes[idest])

                    stack.stack("CRE "+",".join([acid, actype,
                                                 str(self.rwylat[i]),str(self.rwylon[i]),str(self.rwyhdg[i]),
                                                 "0.0","0.0"]))

                    #wplat,wplon = kwikpos(self.rwylat[i],self.rwylon[i],self.rwyhdg[i],5.0*nm)
                    #stack.stack(acid + " ADDWPT ",wplat," ",wplon)
                    #stack.stack(acid+"LNAV ON")
                    if idest>=0:
                        if self.dest[idest][:4] != "SEGM":
                            stack.stack(acid + " DEST " + self.dest[idest])
                        else:
                            stack.stack(acid + " DEST " + str(self.destlat[idest])
                                    + " " + str(self.destlon[idest]))

                    if self.name[:4] != "SEGM":
                        stack.stack(acid + " ORIG " + self.name)
                    else:
                        stack.stack(acid + " ORIG " + str(self.lat) + " " + str(self.lon))

                    stack.stack(acid + " SPD 250")
                    stack.stack(acid + " ALT FL100")
                    stack.stack(acid + " HDG " + str(self.rwyhdg[i]))

                    stack.stack(acid+" LNAV OFF")

            # Not runway, then define instantly at position with random heading or in case of segment inward heading
            if gennow:
                if not self.incircle:
                    lat,lon = kwikpos(ctrlat,ctrlon,self.segdir,radius)
                    hdg = self.segdir-180
                elif self.type=="seg":
                    lat,lon,brg = getseg(self.name)
                    hdg = (brg+180)%360
                elif self.type=="rwy":
                    lat,lon = self.lat,self.lon
                    hdg     = self.hdg # Runway heading
                else:
                    hdg = random.random()*360.

                if self.startaltmin and self.startaltmax:
                    alt = random.randint(int(self.startaltmin), int(self.startaltmax))
                else:
                    alt = random.randint(200, 300) * 100 * ft

                if self.startspdmin and self.startspdmax:
                    spd = random.randint(int(self.startspdmin), int(self.startspdmax))
                else:
                    spd = random.randint(250, 350)

                alttxt, spdtxt = "FL" + str(int(round(alt / (100 * ft)))), str(spd)

                # Add destination
                if len(self.dest)>0:
                    idest = int(random.random() * len(self.dest))
                    acid = randacname(self.name,self.dest[idest])
                else:
                    acid  = randacname(self.name,self.name)
                    idest = -1

                stack.stack("CRE " + ",".join([acid, random.choice(self.actypes),
                                               str(self.lat), str(self.lon), str(int(hdg%360)),
                                               alttxt,spdtxt]))

                if idest>=0:
                    if self.dest[idest][:4] != "SEGM":
                        stack.stack(acid + " DEST " + self.dest[idest])
                    else:
                        stack.stack(acid + " DEST " + str(self.destlat[idest])+" "+str(self.destlon[idest]))

                if self.name[:4] != "SEGM":
                    stack.stack(acid + " ORIG " + self.name)
                else:
                    stack.stack(acid + " ORIG " + str(self.lat)+" "+str(self.lon))

                if alttxt=="0" and spdtxt =="0":
                    stack.stack(acid+" SPD 250")
                    stack.stack(acid+" ALT FL100")
                else:
                    if idest>=0:
                        if self.desttype[idest] == "seg":
                            lat, lon, hdg = getseg(self.dest[idest])
                            brg, dist = kwikqdrdist(self.lat, self.lon, lat, lon)
                            stack.stack(acid + " HDG " + str(brg))
                        else:
                            stack.stack(acid + " LNAV ON")
                            #stack.stack(acid + " VNAV ON")

class Drain():
    # Define a drain: destination within area, source outside
    def __init__(self,name,cmd,cmdargs):
        self.name    = name.upper()
        self.tupdate = sim.simt-999.
        self.flow    = 0
        self.incircle = True
        self.segdir  = None  # Segment direction in degrees

        # Is location a circle segment?
        if self.name[:4]=="SEGM":
            self.type = "seg"
            self.lat,self.lon,brg = getseg(self.name) # For SEGMnnn to SEGMnnn for crossing flights optional
            pass

        else:
            success, posobj = txt2pos(name,ctrlat,ctrlon)
            if success:
                # for runway type, get heading as default optional argument for command line
                if posobj.type == "rwy":
                    aptname, rwyname = name.split('/RW')
                    rwyname = rwyname.lstrip('Y')
                    try:
                        rwyhdg = navdb.rwythresholds[aptname][rwyname][2]
                    except:
                        rwyhdg = None
                        pass
                else:
                    rwyhdg = None


                self.lat,self.lon = posobj.lat, posobj.lon
                self.type = posobj.type

                # If outside circle change lat,lon to edge of circle
                self.incircle = incircle(self.lat,self.lon)
                if not self.incircle:
                    self.type = "seg"
                    self.segdir,dist = qdrdist(ctrlat,ctrlon,posobj.lat,posobj.lon)
                    segname = "SEGM"+str(int(round(self.segdir)))
                    self.lat,self.lon,hdg = getseg(segname)

            else:
                print("ERROR: trafgen class Drain called for "+name+". Position not found")
                self.lat,self.lon = 0.0,0.0


        # Aircraft types
        self.actypes = ["*"]

        # Runways
        self.runways    = [] # name
        self.rwylat     = []
        self.rwylon     = []
        self.rwyhdg     = []
        self.rwyline    = [] # number of aircraft waiting in line

        #Origins
        self.orig        = []
        self.origlat     = []
        self.origlon     = []
        self.origtype    = [] # "apt","wpt","seg"
        self.orighdg     = []  # if orig is a runway (for flights within FIR) or circle segment (drain outside FIR)
        self.origactypes = []   # Types for this originations ([]=use defaults for this drain)
        self.origincirc  = []

        # Limits on start values alt,spd,hdg
        self.startaltmin = None
        self.startaltmax = None
        self.startspdmin = None
        self.startspdmax = None
        self.starthdgmin = None
        self.starthdgmax = None

        #Names of drawing objects runways
        self.polys       = []

        return

    def setrunways(self,cmdargs):
        self.runways = []
        self.rwylat  = []
        self.rwylon  = []
        self.rwyhdg  = []
        for runwayname in cmdargs:
            if runwayname[0] == "R":
                success, rwyposobj = txt2pos(self.name + "/" + runwayname, self.lat, self.lon)
            else:
                success,rwyposobj = txt2pos(self.name+"/RW"+runwayname,self.lat,self.lon)
            if success:
                self.runways.append(runwayname)
                self.rwylat.append(rwyposobj.lat)
                self.rwylon.append(rwyposobj.lon)
                rwyname = runwayname.upper().lstrip('RWY').lstrip("RW")
                try:
                    self.rwyhdg.append(navdb.rwythresholds[self.name][rwyname][2])
                except:
                    success = False

    def addorig(self,cmdargs):
        # Add origin with a given aircraft types
        origname = cmdargs[0]
        try:
            freq = int(cmdargs[1])
        except:
            freq = 1

        # Get a/c types frequency list, if given
        origactypes = []
        if len(cmdargs) > 2:  # also types are given for this origin
            for c in cmdargs[2:]:
                if c.count(":") > 0:
                    actype, typefreq = c.split(":")
                else:
                    actype = c
                    typefreq = "1"

                for f in range(int(typefreq)):
                    origactypes.append(actype)

        if not origname[:4]=="SEGM":
            success,posobj = txt2pos(origname,self.lat,self.lon)
            if success:
                incirc = incircle(posobj.lat, posobj.lon)
                if posobj.type == "rwy":
                    for i in range(freq):
                        self.orig.append(origname)
                        self.origlat.append(posobj.lat)
                        self.origlon.append(posobj.lon)
                        self.orighdg.append(None)
                        self.origtype.append(posobj.type)
                        self.origactypes.append(origactypes)
                        self.origincirc.append(incirc)

                else:
                    for i in range(freq):
                        self.orig.append(origname)
                        self.origlat.append(posobj.lat)
                        self.origlon.append(posobj.lon)
                        self.orighdg.append(None)
                        self.origtype.append(posobj.type)
                        self.origactypes.append(origactypes)
                        self.origincirc.append(incirc)
            else:
                for i in range(freq):
                    name = "SEGM" + str(int(random.random() * 360.))
                    lat, lon, hdg = getseg(name)
                    self.orig.append(name)
                    self.origlat.append(lat)
                    self.origlon.append(lon)
                    self.orighdg.append(None)
                    self.origtype.append("seg")
                    self.origactypes.append(origactypes)
                    self.origincirc.append(False)
        else:
            # Segment as origin, bearing from center = heading
            lat,lon,hdg = getseg(origname)
            self.orig.append(origname)
            self.origlat.append(lat)
            self.origlon.append(lon)
            self.orighdg.append(hdg)
            self.origtype.append("seg")
            self.origactypes.append(origactypes)
            self.origincirc.append(incircle(lat,lon))
        return True

    def setalt(self,cmdargs):
        if len(cmdargs)==1:
            alt = txt2alt(cmdargs[0])
            self.startaltmin = alt
            self.startaltmax = alt
        elif len(cmdargs)>1:
            alt0,alt1 = txt2alt(cmdargs[0]),txt2alt(cmdargs[1])
            self.startaltmin = min(alt0,alt1)
            self.startaltmax = max(alt0,alt1)
        else:
            stack.stack("ECHO "+self.name+" ALT "+str(self.startaltmin)+" "+str(self.startaltmax))

    def setspd(self,cmdargs):
        if len(cmdargs)==1:
            spd = txt2spd(cmdargs[0])
            self.startspdmin = spd
            self.startapdmax = spd
        elif len(cmdargs)>1:
            spd0,spd1 = txt2spd(cmdargs[0]),txt2spd(cmdargs[1])
            self.startspdmin = min(spd0,spd1)
            self.startspdmax = max(spd0,spd1)
        else:
            stack.stack("ECHO "+self.name+" SPD "+str(self.startaltmin)+" "+str(self.startaltmax))

    def sethdg(self,cmdargs):
        if len(cmdargs)==1:
            hdg = float(cmdargs[0])
            self.starthdgmin = hdg
            self.starthdgmax = hdg
        elif len(cmdargs)>1:
            hdg0,hdg1 = float(cmdargs[0]),float(cmdargs[1])
            hdg0,hdg1 = min(hdg0,hdg1),max(hdg0,hdg1)
            if hdg1-hdg0>180.:
                hdg0,hdg1 = hdg1-360,hdg0
            self.starthdgmin = hdg0
            self.starthdgmax = hdg1
        else:
            stack.stack("ECHO "+self.name+" HDG "+str(self.starthdgmin)+" "+str(self.starthdgmax))



    def setflow(self,flowtxt):
        try:
            self.flow = float(flowtxt) #in a/c per hour, also starts flow as it by default zero
        except:
            return False
        return True

    def addactypes(self,actypelist):
        self.actypes = self.actypes + makefreqlist(actypelist)

    def update(self,gain):
        # Time step update of drain

        # Get time step
        dt = sim.simt - self.tupdate
        self.tupdate = sim.simt

        # Time for a new aircraft?
        if dt>0.0:

            # Calculate probability of a geberate occurring with flow
            chances = 1.0-gain*self.flow*dt/3600. #flow is in a/c per hour=360 seconds
            if random.random() >= chances:
                # Calculate starting position using origin
                if len(self.orig)>0:
                    # Add origin
                    iorig = int(random.random() * len(self.orig))
                else:
                    iorig = -1

                if iorig>=0:
                    incirc = self.origincirc[iorig]
                    lat,lon = self.origlat[iorig],self.origlon[iorig]
                    hdg,dist = qdrdist(lat,lon,self.lat,self.lon)
                else:
                    print("Warning update drain",self.name,"called with no origins present!")
                    hdg = random.random()*360.
                    print("using random segment",int(hdg+180)%360)

                    incirc = False

                if not incirc:
                    lat,lon = kwikpos(ctrlat,ctrlon,(hdg+180)%360,radius)
                elif self.origtype=="seg":
                    lat,lon,brg = getseg(self.name)
                    hdg = (brg+180)%360
                else:
                    hdg = random.random()*360.

                if incirc and (self.origtype[iorig]=="apt" or self.origtype[iorig]=="rwy"):
                    alttxt,spdtxt = str(0),str(0)
                else:
                    if self.startaltmin and self.startaltmax:
                        alt = random.randint(int(self.startaltmin), int(self.startaltmax))
                    else:
                        alt = random.randint(200,300)*100*ft

                    if self.startspdmin and self.startspdmax:
                        spd = random.randint(int(self.startspdmin), int(self.startspdmax))
                    else:
                        spd = random.randint(250,350)

                    alttxt,spdtxt = "FL"+str(int(round(alt/(100*ft)))), str(spd)

                if iorig>=0:
                    acid = randacname(self.orig[iorig], self.name)
                else:
                    acid = randacname("LFPG", self.name)

                if len(self.origactypes)>0:
                    actype = random.choice(self.origactypes[iorig])
                else:
                    actype = random.choice(self.actypes)

                stack.stack("CRE " + ",".join([acid,actype,str(lat), str(lon),
                                               str(int(hdg%360)),alttxt,spdtxt]))
                if iorig>=0:
                    if self.orig[iorig][:4]!="SEGM":
                        stack.stack(acid + " ORIG " + self.orig[iorig])
                    else:
                        stack.stack(acid + " ORIG " + str(self.origlat[iorig]) + " " +\
                                     str(self.origlon[iorig]))
                if not (self.name[:4]=="SEGM"):
                    stack.stack(acid + " DEST " + self.name)
                else:
                    stack.stack(acid + " ADDWPT " + str(self.lat) + " " + str(self.lon))

                if alttxt=="0" and spdtxt =="0":
                    stack.stack(acid+" SPD 250")
                    stack.stack(acid+" ALT FL100")
                    #stack.stack(acid+" LNAV ON")
                else:
                    stack.stack(acid + " LNAV ON")
                    #stack.stack(acid + " VNAV ON") ATC discretion

def randacname(orig,dest):
    companies = 70*["KLM"]+30*["TRA"]+10*["**"]+["PH"]
    company = random.choice(companies)
    if dest[:2]=="EH" and orig[:2]=="EH":
        company = "PH"

    if company=="**":
        company = chr(ord("A") + int(random.random()*26))+ \
                  chr(ord("A") + int(random.random()*26))

    #if orig=="EHKD"

    # Make flight number or Dutch call sign for VFR traffic
    firstx =  True
    while firstx or (acname in traf.id):
        if not (company=="PH"):
            fltnr = str(int(random.random()*900+100))
        else:
            fltnr = "-" + chr(ord("A") + int(random.random()*26))+ \
                chr(ord("A") + int(random.random() * 26)) + \
                chr(ord("A") + int(random.random() * 26))

        acname = company+fltnr
        firstx = False

    return company+fltnr

def makefreqlist(txtlist):
    # Expand freqlist
    # Translate  arguments ["KL:3","HV:1","PH","MP:5"] into a pick list with the relative frequency:
    #           ["KL","KL","KL","HV","PH","MP","MP","MP","MP","MP"]
    lst = []
    for item in txtlist:
        if item.count(":"):
            itemtxt,freq = item.strip().split(":")
            for i in range(int(freq)):
                lst.append(itemtxt)
        else:
            lst.append(item)
    return lst


def getseg(txt):

    # Get a random position on the segment with the heading inward
    # SEGMnnn is segment in direction nnn degrees from center circle

    brg = float(txt[4:])

    lat,lon = kwikpos(ctrlat,ctrlon,brg,radius)

    return lat,lon,brg


def checkactype(curtype,dist,alltypes):
    # Use lookup table/size/weight to match flight distance with a suitable aircraft type (if available)
    # To avoid B744s flying to Eelde and PA28 to New York
    newtype = curtype

    # Temporary quick fix
    # TBD
    if curtype[0]=="A" or curtype[0]=="B": # Likely Airbus or Boeing? => no small distance < 300 nm
        if dist<300:
            # Try to find another one maxiumum 25 times
            n = 25
            while n>0 and newtype[0] in ["A","B"]:
                n = n - 1
                newtype = random.choice(alltypes)
    else:
        if dist > 500:
            # Try to find another one maxiumum 25 times
            n = 25
            while n > 0 and not( newtype[0] in ["A", "B"]):
                n = n - 1
                newtype = random.choice(alltypes)

    return newtype

def incircle(lat,lon): # Check whether position is inside current circle definition
    if not swcircle:
        return True
    else:
        dist = latlondist(ctrlat,ctrlon,lat,lon)/nm
        return dist<=radius


def crosscircle(xm,ym,r,xa,ya,xb,yb):
    # Find crossing points of line throught a end b and circle around m with radius r
    a1 = xb-xa
    a2 = yb-ya
    b1 = xb-xm
    b2 = yb-ym
    A = a1*a1+a2*a2
    B = 2*a1*b1+2*a2*b2
    C = b1*b1+b2*b2-r*r
    D = B*B-4*A*C
    if D<0:
        return [None,None]
    else:
        VD = sqrt(D)
        lam1 = (-B-VD)/(2*A)
        x1,y1 = b1+lam1*a1+xm , b2+lam1*a2+ym
        lam2 = (-B+VD)/(2*A)
        x2, y2 = b1+lam2*a1+xm , b2+lam2*a2+ym
        return [[x1,y1],[x2,y2]]



""" This module defines a set of standard geographic functions and constants for
    easy use in BlueSky. """
import numpy as np
from math import *

import bluesky as bs


# Constants
nm  = 1852.  # m       1 nautical mile

# Read data for declination switch
decl_read = False
decl_lat_lon = np.array([])


def rwgs84(latd):
    """ Calculate the earths radius with WGS'84 geoid definition
        In:  lat [deg] (latitude)
        Out: R   [m]   (earth radius) """
    lat    = np.radians(latd)
    a      = 6378137.0       # [m] Major semi-axis WGS-84
    b      = 6356752.314245  # [m] Minor semi-axis WGS-84
    coslat = np.cos(lat)
    sinlat = np.sin(lat)

    an     = a * a * coslat
    bn     = b * b * sinlat
    ad     = a * coslat
    bd     = b * sinlat

    # Calculate radius in meters
    r = np.sqrt((an * an + bn * bn) / (ad * ad + bd * bd))

    return r
#------------------------------------------------------------


def rwgs84_matrix(latd):
    """ Calculate the earths radius with WGS'84 geoid definition
        In:  lat [deg] (Vector of latitudes)
        Out: R   [m]   (Vector of radii) """

    lat    = np.radians(latd)
    a      = 6378137.0       # [m] Major semi-axis WGS-84
    b      = 6356752.314245  # [m] Minor semi-axis WGS-84
    coslat = np.cos(lat)
    sinlat = np.sin(lat)
    an     = a * a * coslat
    bn     = b * b * sinlat
    ad     = a * coslat
    bd     = b * sinlat

    anan   = np.multiply(an, an)
    bnbn   = np.multiply(bn, bn)
    adad   = np.multiply(ad, ad)
    bdbd   = np.multiply(bd, bd)
    # Calculate radius in meters
    r      = np.sqrt(np.divide(anan + bnbn, adad + bdbd))

    return r


def qdrdist(latd1, lond1, latd2, lond2):
    """ Calculate bearing and distance, using WGS'84
        In:
            latd1,lond1 en latd2, lond2 [deg] :positions 1 & 2
        Out:
            qdr [deg] = heading from 1 to 2
            d [nm]    = distance from 1 to 2 in nm """

    # Haversine with average radius for direction

    # Check for hemisphere crossing,
    # when simple average would not work

    # res1 for same hemisphere
    res1 = rwgs84(0.5 * (latd1 + latd2))

    # res2 :different hemisphere
    a    = 6378137.0       # [m] Major semi-axis WGS-84
    r1   = rwgs84(latd1)
    r2   = rwgs84(latd2)
    res2 = 0.5 * (abs(latd1) * (r1 + a) + abs(latd2) * (r2 + a)) / \
        (np.maximum(0.000001,abs(latd1) + abs(latd2)))

    # Condition
    sw   = (latd1 * latd2 >= 0.)

    r    = sw * res1 + (1 - sw) * res2

    # Convert to radians
    lat1 = np.radians(latd1)
    lon1 = np.radians(lond1)
    lat2 = np.radians(latd2)
    lon2 = np.radians(lond2)

    #root = sin1 * sin1 + coslat1 * coslat2 * sin2 * sin2
    #d    =  2.0 * r * np.arctan2(np.sqrt(root) , np.sqrt(1.0 - root))
    # d =2.*r*np.arcsin(np.sqrt(sin1*sin1 + coslat1*coslat2*sin2*sin2))

    # Corrected to avoid "nan" at westward direction
    d = r*np.arccos(np.cos(lat1)*np.cos(lat2)*np.cos(lon2-lon1) + \
                 np.sin(lat1)*np.sin(lat2))

    # Bearing from Ref. http://www.movable-type.co.uk/scripts/latlong.html

    # sin1 = np.sin(0.5 * (lat2 - lat1))
    # sin2 = np.sin(0.5 * (lon2 - lon1))

    coslat1 = np.cos(lat1)
    coslat2 = np.cos(lat2)

    qdr = np.degrees(np.arctan2(np.sin(lon2 - lon1) * coslat2,
                                coslat1 * np.sin(lat2) -
                                np.sin(lat1) * coslat2 * np.cos(lon2 - lon1)))

    return qdr, d/nm


def qdrdist_matrix(lat1, lon1, lat2, lon2):
    """ Calculate bearing and distance vectors, using WGS'84
        In:
            latd1,lond1 en latd2, lond2 [deg] :positions 1 & 2 (vectors)
        Out:
            qdr [deg] = heading from 1 to 2 (matrix)
            d [nm]    = distance from 1 to 2 in nm (matrix) """
    prodla =  lat1.T * lat2
    condition = prodla < 0

    r = np.zeros(prodla.shape)
    r = np.where(condition, r, rwgs84_matrix(lat1.T + lat2))

    a = 6378137.0

    r = np.where(np.invert(condition), r, (np.divide(np.multiply
      (0.5, ((np.multiply(abs(lat1), (rwgs84_matrix(lat1)+a))).T +
         np.multiply(abs(lat2), (rwgs84_matrix(lat2)+a)))),
            (abs(lat1)).T+(abs(lat2)+(lat1 == 0.)*0.000001))))  # different hemisphere

    diff_lat = lat2-lat1.T
    diff_lon = lon2-lon1.T

    sin1 = (np.radians(diff_lat))
    sin2 = (np.radians(diff_lon))

    sinlat1 = np.sin(np.radians(lat1))
    sinlat2 = np.sin(np.radians(lat2))
    coslat1 = np.cos(np.radians(lat1))
    coslat2 = np.cos(np.radians(lat2))

    sin21 = np.mat(np.sin(sin2))
    cos21 = np.mat(np.cos(sin2))
    y = np.multiply(sin21, coslat2)

    x1 = np.multiply(coslat1.T, sinlat2)

    x2 = np.multiply(sinlat1.T, coslat2)
    x3 = np.multiply(x2, cos21)
    x = x1-x3

    qdr = np.degrees(np.arctan2(y, x))

    sin10 = np.mat(np.abs(np.sin(sin1/2.)))
    sin20 = np.mat(np.abs(np.sin(sin2/2.)))
    sin1sin1 = np.multiply(sin10, sin10)
    sin2sin2 = np.multiply(sin20, sin20)
    sqrt = sin1sin1 + np.multiply((coslat1.T * coslat2), sin2sin2)
    dist_c = np.multiply(2., np.arctan2(np.sqrt(sqrt), np.sqrt(1-sqrt)))
    dist = np.multiply(r/nm, dist_c)
    #    dist = np.multiply(2.*r, np.arcsin(sqrt))

    return qdr, dist


def latlondist(latd1, lond1, latd2, lond2):
    """ Calculates only distance using haversine notation of the same formulae and average r from wgs'84
        Input:
              two lat/lon positions in degrees
        Out:
              distance in meters !!!! """

    # Haversine with average radius

    # Check for hemisphere crossing,
    # when simple average would not work

    # res1 for same hemisphere
    res1 = rwgs84(0.5*(latd1+latd2))

    # res2 :different hemisphere
    a = 6378137.0       # [m] Major semi-axis WGS-84
    r1 = rwgs84(latd1)
    r2 = rwgs84(latd2)
    res2  = 0.5*(abs(latd1)*(r1+a) + abs(latd2)*(r2+a)) / \
        (abs(latd1)+abs(latd2))

    # Condition
    sw = (latd1*latd2 >= 0.)

    r = sw*res1+(1-sw)*res2

    # Convert to radians
    lat1 = np.radians(latd1)
    lon1 = np.radians(lond1)
    lat2 = np.radians(latd2)
    lon2 = np.radians(lond2)

    sin1 = np.sin(0.5*(lat2-lat1))
    sin2 = np.sin(0.5*(lon2-lon1))

    coslat1 = np.cos(lat1)
    coslat2 = np.cos(lat2)

    root = sin1*sin1 + coslat1*coslat2*sin2*sin2
    d =  2.*r*np.arctan2(np.sqrt(root), np.sqrt(1.-root))

    #    d =2*r*np.arcsin(np.sqrt(sin1*sin1 + coslat1*coslat2*sin2*sin2))
    return d


def latlondist_matrix(lat1, lon1, lat2, lon2):
    """ Calculates distance using haversine formulae and avaerage r from wgs'84
        Input:
              two lat/lon position vectors in degrees
        Out:
              distance vector in meters !!!! """
    prodla =  lat1.T*lat2
    condition = prodla < 0

    r = np.zeros(len(prodla))
    r = np.where(condition, r, rwgs84_matrix(lat1.T+lat2))

    a = 6378137.0
    r = np.where(np.invert(condition), r, (np.divide(np.multiply(0.5,
        ((np.multiply(abs(lat1), (rwgs84_matrix(lat1)+a))).T +
            np.multiply(abs(lat2), (rwgs84_matrix(lat2)+a)))),
            (abs(lat1)).T+(abs(lat2)))))  # different hemisphere

    diff_lat = lat2-lat1.T
    diff_lon = lon2-lon1.T

    sin1 = (np.radians(diff_lat))
    sin2 = (np.radians(diff_lon))

    coslat1 = np.cos(np.radians(lat1))
    coslat2 = np.cos(np.radians(lat2))

    sin10 = np.mat(np.sin(sin1/2))
    sin20 = np.mat(np.sin(sin2/2))
    sin1sin1 =  np.multiply(sin10, sin10)
    sin2sin2 =  np.multiply(sin20, sin20)
    root =  sin1sin1+np.multiply((coslat1.T*coslat2), sin2sin2)

    #    dist = np.multiply(2.*r, np.arcsin(sqrt))
    dist_c =  np.multiply(2, np.arctan2(np.sqrt(root), np.sqrt(1.-root)))
    dist = np.multiply(r/nm, dist_c)

    return dist


def wgsg(latd):
    """ Gravity acceleration at a given latitude according to WGS'84 """
    geq = 9.7803   # m/s2 g at equator
    e2 = 6.694e-3  # eccentricity
    k  = 0.001932  # derived from flattening f, 1/f = 298.257223563

    sinlat = np.sin(np.radians(latd))
    g = geq*(1.0 + k*sinlat*sinlat) / np.sqrt(1.0 - e2*sinlat*sinlat)

    return g


def qdrpos(latd1, lond1, qdr, dist):
    """ Calculate vector with positions from vectors of reference position,
        bearing and distance.
        In:
             latd1,lond1  [deg]   ref position(s)
             qdr          [deg]   bearing (vector) from 1 to 2
             dist         [nm]    distance (vector) between 1 and 2
        Out:
             latd2,lond2 (IN DEGREES!)
        Ref for qdrpos: http://www.movable-type.co.uk/scripts/latlong.html """

    # Unit conversion
    R = rwgs84(latd1)/nm
    lat1 = np.radians(latd1)
    lon1 = np.radians(lond1)

    # Calculate new position
    lat2 = np.arcsin(np.sin(lat1)*np.cos(dist/R) +
                     np.cos(lat1)*np.sin(dist/R)*np.cos(np.radians(qdr)))

    lon2 = lon1 + np.arctan2(np.sin(np.radians(qdr))*np.sin(dist/R)*np.cos(lat1),
                             np.cos(dist/R) - np.sin(lat1)*np.sin(lat2))
    return np.degrees(lat2), np.degrees(lon2)


def kwikdist(lata, lona, latb, lonb):
    """
    Quick and dirty dist [nm]
    In:
        lat/lon, lat/lon [deg]
    Out:
        dist [nm]
    """

    re      = 6371000.  # radius earth [m]
    dlat    = np.radians(latb - lata)
    dlon    = np.radians(((lonb - lona)+180)%360-180)
    cavelat = np.cos(np.radians(lata + latb) * 0.5)

    dangle  = np.sqrt(dlat * dlat + dlon * dlon * cavelat * cavelat)
    dist    = re * dangle / nm

    return dist


def kwikdist_matrix(lata, lona, latb, lonb):
    """
    Quick and dirty dist [nm]
    In:
        lat/lon, lat/lon vectors [deg]
    Out:
        dist vector [nm]
    """

    re      = 6371000.  # readius earth [m]
    dlat    = np.radians(latb - lata.T)
    dlon    = np.radians(((lonb - lona.T)+180)%360-180)
    cavelat = np.cos(np.radians(lata + latb.T) * 0.5)

    dangle  = np.sqrt(np.multiply(dlat, dlat) +
                      np.multiply(np.multiply(dlon, dlon),
                                  np.multiply(cavelat, cavelat)))
    dist    = re * dangle / nm

    return dist


def kwikqdrdist(lata, lona, latb, lonb):
    """Gives quick and dirty qdr[deg] and dist [nm]
       from lat/lon. (note: does not work well close to poles)"""

    re      = 6371000.  # radius earth [m]
    dlat    = np.radians(latb - lata)
    dlon    = np.radians(((lonb - lona)+180)%360-180)
    cavelat = np.cos(np.radians(lata + latb) * 0.5)

    dangle  = np.sqrt(dlat * dlat + dlon * dlon * cavelat * cavelat)
    dist    = re * dangle / nm

    qdr     = np.degrees(np.arctan2(dlon * cavelat, dlat)) % 360.

    return qdr, dist


def kwikqdrdist_matrix(lata, lona, latb, lonb):
    """Gives quick and dirty qdr[deg] and dist [nm] matrices
       from lat/lon vectors. (note: does not work well close to poles)"""

    re      = 6371000.  # radius earth [m]
    dlat    = np.radians(latb - lata.T)
    dlon    = np.radians(((lonb - lona.T)+ 180) % 360 - 180)
    cavelat = np.cos(np.radians(latb + lata.T) * 0.5)

    dangle  = np.sqrt(np.multiply(dlat, dlat) +
                      np.multiply(np.multiply(dlon, dlon),
                                  np.multiply(cavelat, cavelat)))
    dist    = re * dangle / nm

    qdr     = np.degrees(np.arctan2(np.multiply(dlon, cavelat), dlat)) % 360.

    return qdr, dist

def kwikpos(latd1, lond1, qdr, dist):
    """ Fast, but quick and dirty, position calculation from vectors of reference position,
        bearing and distance using flat earth approximation
        In:
             latd1,lond1  [deg]   ref position(s)
             qdr          [deg]   bearing (vector) from 1 to 2
             dist         [nm]    distance (vector) between 1 and 2
        Out:
             latd2,lond2 [deg]
        Use for flat earth purposes e.g. flat display"""

    dx = dist*np.sin(np.radians(qdr))
    dy = dist*np.cos(np.radians(qdr))
    dlat = dy/60.
    dlon = dx/(np.maximum(0.01,60.*np.cos(np.radians(latd1))))
    latd2 = latd1 + dlat
    lond2 = ((lond1 + dlon)+180)%360-180

    return latd2,lond2

def magdec(latd, lond):
    """
    Gives magnetic declination (also called magnetic variation) at given
    position, interpolated from an external data table. The interpolation is
    done using an object of scipy.interpolate.RectSphereBivariateSpline
    interpo_dec, which is generated by the function init_interpo_dec() defined
    in the same module (geo.py). The interpo_dec object rvaluates the magnetic
    declination at any latitude and longitude (latd, lond).
    The function magdec() first checks if the object interpo_dec exists and is
    an object of scipy.interpolate.RectSphereBivariateSpline. If not found or
    not be the case, as happends when magdec is initially called, the function
    init_interpo_dec() will be called.
    The arguments of interpo_dec.ev() are 1-D arrays of latitudes and
    longitudes in radians, with latitude ranging from 0 to pi and longitude
    ranging from 0 to 2pi.
    In:
         latd, lond  [deg]  Position at which the magnetic declination is
                            evaluated (floats)
    Out:
         d_hdg       [deg]  Magnetic declination, the angle of difference
                            between true North and magnetic North. For instance,
                            if the declination at a certain point were 10 deg W
                            (10 deg), then a compass at that location pointing
                            north (magnetic) would actually align 10 deg W of
                            true North. True North would be 10 deg E relative to
                            the magnetic North direction given by the compass.
                            Declination varies with location and slowly changes
                            in time. Referenced from
            https://www.ngdc.noaa.gov/geomag/calculators/help/igrfgridHelp.html
                            In short, magnetic heading = true heading - d_hdg,
                            (Reminder MTV : M = T - V)
                            or,       true heading = magnetic heading + d_hdg.
    Created by  : Yaofu Zhou
    Modified by J.M. Hoekstra
    Reason: Segmentation fault caused by Scipy's BiVariateSpline interpolation
    for some data on some machines, so it was changed to linear interpolation.
    Difference in methods has been inspected: it is way less than the inaccuracy
    of the actual data. Axes were regularly spaced at one degree. The direct
    manual linear interpolation also 6 x times faster.
    """
    global decl_read
    if not decl_read:
        initdecl_data()
        decl_read = True

    # Use fact that whole degrees are used as ticks on both lat & lon axis
    i_lat = min(max(0,int(90.-latd)),180)
    f_lat = (90.-latd) - int(90.-latd)
    i_lon = min(max(0,int(lond + 180)),360)
    f_lon = lond+180. - int(lond+180)

    # 2D linear interpolation
    declon0 = decl_lat_lon[i_lat,i_lon]*(1.-f_lat) + \
               f_lat*decl_lat_lon[min(180,i_lat+1),i_lon]
    declon1 = decl_lat_lon[i_lat,i_lon+1]*(1.-f_lat) + \
               f_lat*decl_lat_lon[min(180,i_lat+1),min(i_lon+1,360)]

    d_hdg = declon0*(1.-f_lon) + f_lon*declon1

    return d_hdg

def initdecl_data():
    """
    Called by Init
    Read magnetic declination (also called magnetic variation) datafile
    based on the data table calculated from the NOAA webpage
    https://www.ngdc.noaa.gov/geomag/calculators/magcalc.shtml#igrfgrid
    with the following input:
        Southern most lat:  90 S
        Northern most lat:  90 N
        Lat Step Size:      1.0
        Western most long:  180 W
        Eastern most long:  179 E
        Lon Step Size:      1.0
        Elevation:          Mean sea level 0 Feet
        Magnetic component: Declination
        Model:              WMM (2019-2024)
        Start Date:         2020 09 20
        End Date:           2020 09 20
        Step size:          1.0
        Result format:      CSV
    The grid size can be adjusted but the (1 deg by 1 deg) size should suffice
    for practical purpose, as long as the the grids cover the entire Earth
    surface. The interpolation is performed at sea-level, but no significant
    difference would be noticed up to FL600 or beyond.
    See docstring of geo.magdec() for more information.
    Based on original version created by  : Yaofu Zhou
    Modified to read at init and use linear interpolation by J.M. Hoekstra"""

    #    Columns:
    #     (1) Date in decimal years
    #     (2) Latitude in decimal Degrees
    #     (3) Longitude in decimal Degrees
    #     (4) Elevation in km Mean Sea Level
    #     (5) Declination in Degree
    #     (6) Declination_sv in Degree
    #     (7) Declination_uncertainty in Degree
    #
    # lat : 89 ... -90
    # Lon: -180 ... 179
    global decl_read, decl_lat_lon
    dec_table = np.genfromtxt(bs.resource(bs.settings.navdata_path) / 'geo_declination_data.csv',
                              comments='#',delimiter=",")

    decl = dec_table[:,4]

    #          <----lon ---->
    #   lat1    ..  ..   ..  ..
    #   lat2    .. ..

    # Correct and adapt table
    decl_lat_lon = decl.reshape((180,360))
    # For intpol, add lat = 90 row (same as at 89 degrees for interpolation)
    # even though it is in blackout zone around pole
    decl_lat_lon = np.vstack((decl_lat_lon[0:1,:],decl_lat_lon))
    # Add lon = +180 column
    decl_lat_lon = np.hstack((decl_lat_lon, decl_lat_lon[:,0:1]))

    # Result is table 181 rows x 361 columns table for
    # lat = 90 ... -90 (rows)
    # lon = -180 ... 180 (columns)
    decl_read = True

    return decl_lat_lon

# Command MAGVAR to get magnetic variation at position lat,lon
def magdeccmd(latdeg,londeg):
    ''' MAGVAR Get magnetic variation at position lat/lon. '''
    return True,"Magnetic variation at "+str(latdeg)+","+str(londeg) + \
                " = "+str(magdec(latdeg,londeg))+" deg"



from pathlib import Path
import subprocess
import traceback
import math

import bluesky as bs
from bluesky.core.signal import Signal
from bluesky.stack.stackbase import Stack, forward, stack
from bluesky.stack.cmdparser import Command, command, commandgroup
from bluesky.stack import argparser


# Globals
_sig_echo = Signal('echo')


def init():
    ''' client-side stack initialisation. '''
    # Display Help text on start of program
    stack("ECHO BlueSky Console Window: Enter HELP or ? for info.\n"
          "Or select IC to Open a scenario file.")


def process():
    ''' Client-side stack processing. '''
    # Process stack of commands
    for cmdline in Stack.commands():
        success = True
        echotext = ''
        echoflags = bs.BS_OK

        # Get first argument from command line and check if it's a command
        cmd, argstring = argparser.getnextarg(cmdline)
        cmdu = cmd.upper()
        cmdobj = Command.cmddict.get(cmdu)

        # Proceed if a command object was found
        if cmdobj:
            try:
                # Call the command, passing the argument string
                success, echotext = cmdobj(argstring)
                if not success:
                    if not argstring:
                        echotext = echotext or cmdobj.brieftext()
                    else:
                        echoflags = bs.BS_FUNERR
                        echotext = f'Syntax error: {echotext or cmdobj.brieftext()}'

            except Exception as e:
                success = False
                echoflags = bs.BS_ARGERR
                header = '' if not argstring else e.args[0] if e.args else 'Argument error.'
                echotext = f'{header}\nUsage:\n{cmdobj.brieftext()}'
                traceback.print_exc()

        # ----------------------------------------------------------------------
        # ZOOM command (or use ++++  or --  to zoom in or out)
        # ----------------------------------------------------------------------
        elif cmdu[0] in ("+", "=", "-"):
            # = equals + (same key)
            nplus = cmdu.count("+") + cmdu.count("=")
            nmin = cmdu.count("-")
            fac = math.sqrt(2) ** (nplus - nmin)
            cmdu = 'ZOOM'
            cmdobj = Command.cmddict.get(cmdu)
            if cmdobj:
                cmdobj(f'IN {fac}')

        elif Stack.sender_id is None:
            # If sender_id is None, this stack command originated from the gui. Send it on to the sim
            forward()
        # -------------------------------------------------------------------
        # Command not found
        # -------------------------------------------------------------------
        else:
            success = False
            echoflags = bs.BS_CMDERR
            if not argstring:
                echotext = f'Unknown command or aircraft: {cmd}'
            else:
                echotext = f'Unknown command: {cmd}'

        # Always return on command
        if echotext:
            echo(echotext, echoflags, Stack.sender_id)

    # Clear the processed commands
    Stack.cmdstack.clear()


@command(annotations='string')
def echo(text='', flags=0, sender_id=None):
        ''' Echo

            Clien-side implementation of ECHO emits the same
            signal as the one triggered by incoming echo
            messages.    
        '''
        _sig_echo.emit(text, flags, sender_id)


@commandgroup(name='HELP', aliases=('?',))
def showhelp(cmd: 'txt' = '', subcmd: 'txt' = ''):
    """ HELP: Display general help text or help text for a specific command,
        or dump command reference in file when command is >filename.

        Arguments:
        - cmd: Argument can refer to:
            - Command name to display help for. 
            - Call HELP >filename to generate a CSV file with help text for all commands.
            - Call HELP PDF to view a pdf containing information on all stack commands.

        To get more detailed information on a command type DOC cmd.
    """

    # Check if help is asked for a specific command
    cmdobj = Command.cmddict.get(cmd or 'HELP')
    if cmdobj:
        return True, cmdobj.helptext(subcmd)

    # If command is not a known Client command pass the help request on to the sim
    forward(target_id=bs.net.act_id)


@showhelp.subcommand
def pdf():
    ''' Open a pdf file with BlueSky command help text. '''
    pdfhelp = Path("BLUESKY-COMMAND-TABLE.pdf")
    if (Path('docs') / pdfhelp).is_file():
        try:
            subprocess.Popen(pdfhelp, shell=True, cwd='docs')
        except:
            return "Opening " + pdfhelp.as_posix() + " failed."
    else:
        return pdfhelp.as_posix() + "does not exist."

    return "Pdf window opened"


@command
def makedoc():
    ''' MAKEDOC: Make markdown templates for all stack functions
        that don't have a doc page yet.
    '''
    tmp = Path('tmp')
    if not tmp.is_dir():
        tmp.mkdir()
    # Get unique set of commands
    cmdobjs = set(Command.cmddict.values())
    for o in cmdobjs:
        if not bs.resource(f"html/{o.name}.html").is_file():
            with open(tmp / f"{o.name.lower()}.md", "w") as f:
                f.write(
                    f"# {o.name}: {o.name.capitalize()}\n"
                    + o.help
                    + "\n\n"
                    + "**Usage:**\n\n"
                    + f"    {o.brief}\n\n"
                    + "**Arguments:**\n\n"
                )
                if not o.params:
                    f.write("This command has no arguments.\n\n")
                else:
                    f.write(
                        "|Name|Type|Optional|Description\n"
                        + "|--------|------|---|---------------------------------------------------\n"
                    )
                    for arg in o.params:
                        f.write(str(arg).replace(':', '|') + f" |{arg.hasdefault()}|\n")
                f.write("\n[[Back to command reference.|Command Reference]]\n")
    # To also get all of the sim stack commands, forward to sim
    forward()


"""BlueSky ADS-B datafeed plugin. Reads the feed from a Mode-S Beast server,
and visualizes traffic in BlueSky."""

import time

from bluesky import settings, stack, traf
from bluesky.tools import aero
from bluesky.tools.network import TcpSocket

## Default settings
# Mode-S / ADS-B server hostname/ip, and server port
settings.set_variable_defaults(modeS_host="", modeS_port=0)

# Global data
reader = None


### Initialization function of the adsbfeed plugin.
def init_plugin():
    # Initialize Modesbeast reader
    global reader
    reader = Modesbeast()

    # Configuration parameters
    config = {
        "plugin_name": "DATAFEED",
        "plugin_type": "sim",
        "update_interval": 0.0,
        "preupdate": reader.update,
    }

    stackfunctions = {
        "DATAFEED": [
            "DATAFEED [ON/OFF]",
            "[onoff]",
            reader.toggle,
            "Select an ADS-B data source for traffic",
        ]
    }

    # init_plugin() should always return these two dicts.
    return config, stackfunctions


class Modesbeast(TcpSocket):
    def __init__(self):
        super().__init__()
        self.acpool = {}
        self.buffer = b""
        self.default_ac_mdl = "B738"

    def processData(self, data):
        self.buffer += data

        if len(self.buffer) > 2048:
            # process the buffer until the last divider <esc> 0x1a
            # then, reset the buffer with the remainder

            bfdata = self.buffer
            n = (len(bfdata) - 1) - bfdata[::-1].index(0x1A)
            data = bfdata[: n - 1]
            self.buffer = self.buffer[n:]

            messages = self.read_mode_s(data)

            if not messages:
                return

            for msg, ts in messages:
                self.read_message(msg, ts)
        return

    def read_mode_s(self, data):
        """
        <esc> "1" : 6 byte MLAT timestamp, 1 byte signal level,
            2 byte Mode-AC
        <esc> "2" : 6 byte MLAT timestamp, 1 byte signal level,
            7 byte Mode-S short frame
        <esc> "3" : 6 byte MLAT timestamp, 1 byte signal level,
            14 byte Mode-S long frame
        <esc> "4" : 6 byte MLAT timestamp, status data, DIP switch
            configuration settings (not on Mode-S Beast classic)
        <esc><esc>: true 0x1a
        <esc> is 0x1a, and "1", "2" and "3" are 0x31, 0x32 and 0x33

        timestamp:
        wiki.modesbeast.com/Radarcape:Firmware_Versions#The_GPS_timestamp
        """

        # split raw data into chunks
        chunks = []
        separator = 0x1A
        piece = []
        for d in data:
            if d == separator:
                # shortest msgs are 11 chars
                if len(piece) > 10:
                    chunks.append(piece)
                piece = []
            piece.append(d)

        # extract messages
        messages = []
        for cnk in chunks:
            msgtype = cnk[1]

            # Mode-S Short Message, 7 byte
            if msgtype == 0x32:
                msg = "".join("%02X" % i for i in cnk[9:16])

            # Mode-S Short Message, 14 byte
            elif msgtype == 0x33:
                msg = "".join("%02X" % i for i in cnk[9:23])

            # Other message tupe
            else:
                continue

            ts = time.time()

            messages.append([msg, ts])
        return messages

    def read_message(self, msg, ts):
        """
        Process ADSB messages
        """

        if len(msg) < 28:
            return

        df = Decoder.get_df(msg)

        if df == 17:
            addr = Decoder.get_icao_addr(msg)
            tc = Decoder.get_tc(msg)

            if tc >= 1 and tc <= 4:
                # aircraft identification
                callsign = Decoder.get_callsign(msg)
                self.update_callsign(addr, callsign)
            if tc >= 9 and tc <= 18:
                # airbone postion frame
                alt = Decoder.get_alt(msg)
                oe = Decoder.get_oe_flag(msg)  # odd or even frame
                cprlat = Decoder.get_cprlat(msg)
                cprlon = Decoder.get_cprlon(msg)
                self.update_cprpos(addr, oe, ts, alt, cprlat, cprlon)
            elif tc == 19:  # airbone velocity frame
                sh = Decoder.get_speed_heading(msg)
                if len(sh) == 2:
                    spd = sh[0]
                    hdg = sh[1]
                    self.update_spd_hdg(addr, spd, hdg)
        return

    def update_cprpos(self, addr, oe, ts, alt, cprlat, cprlon):
        if addr in self.acpool:
            ac = self.acpool[addr]
        else:
            ac = {}

        ac["alt"] = alt
        if oe == "1":  # odd frame cpr position
            ac["cprlat1"] = cprlat
            ac["cprlon1"] = cprlon
            ac["t1"] = ts

        if oe == "0":  # even frame cpr position
            ac["cprlat0"] = cprlat
            ac["cprlon0"] = cprlon
            ac["t0"] = ts

        ac["ts"] = time.time()

        self.acpool[addr] = ac
        return

    def update_spd_hdg(self, addr, spd, hdg):
        if addr in self.acpool:
            ac = self.acpool[addr]
        else:
            ac = {}

        ac["speed"] = spd
        ac["heading"] = hdg
        ac["ts"] = time.time()

        self.acpool[addr] = ac
        return

    def update_callsign(self, addr, callsign):
        if addr not in self.acpool:
            self.acpool[addr] = {}

        self.acpool[addr]["callsign"] = callsign
        return

    def update_all_ac_postition(self):
        keys = ("cprlat0", "cprlat1", "cprlon0", "cprlon1")
        for addr, ac in list(self.acpool.items()):
            # check if all needed keys are in dict
            if set(keys).issubset(ac):
                pos = Decoder.cpr2position(
                    ac["cprlat0"],
                    ac["cprlat1"],
                    ac["cprlon0"],
                    ac["cprlon1"],
                    ac["t0"],
                    ac["t1"],
                )

                # update positions of all aircrafts in the list
                if pos:
                    self.acpool[addr]["lat"] = pos[0]
                    self.acpool[addr]["lon"] = pos[1]
        return

    def stack_all_commands(self):
        """create and stack command"""
        params = ("lat", "lon", "alt", "speed", "heading", "callsign")
        for i, d in list(self.acpool.items()):
            # check if all needed keys are in dict
            if set(params).issubset(d):
                acid = d["callsign"]
                # check is aircraft is already beening displayed
                if traf.id2idx(acid) < 0:
                    mdl = self.default_ac_mdl
                    v = aero.tas2cas(d["speed"], d["alt"] * aero.ft)
                    cmdstr = "CRE %s, %s, %f, %f, %f, %d, %d" % (
                        acid,
                        mdl,
                        d["lat"],
                        d["lon"],
                        d["heading"],
                        d["alt"],
                        v,
                    )
                    stack.stack(cmdstr)
                else:
                    cmdstr = "MOVE %s, %f, %f, %d" % (
                        acid,
                        d["lat"],
                        d["lon"],
                        d["alt"],
                    )
                    stack.stack(cmdstr)

                    cmdstr = "HDG %s, %f" % (acid, d["heading"])
                    stack.stack(cmdstr)

                    v_cas = aero.tas2cas(d["speed"], d["alt"] * aero.ft)
                    cmdstr = "SPD %s, %f" % (acid, v_cas)
                    stack.stack(cmdstr)
        return

    def remove_outdated_ac(self):
        """House keeping, remove old entries (offline > 100s)"""
        for addr, ac in list(self.acpool.items()):
            if "ts" in ac:
                # threshold, remove ac after 90 seconds of no-seen
                if (int(time.time()) - ac["ts"]) > 100:
                    del self.acpool[addr]
                    # remove from sim traffic
                    if "callsign" in ac:
                        stack.stack("DEL %s" % ac["callsign"])
        return

    def debug(self):
        addlist = str.join(", ", self.acpool.keys())
        print(addlist)
        print("")
        print("total count: %d" % len(self.acpool.keys()))
        return

    def update(self):
        if self.isConnected():
            # self.debug()
            self.remove_outdated_ac()
            self.update_all_ac_postition()
            self.stack_all_commands()

    def toggle(self, flag=None):
        if flag is None:
            if self.isConnected():
                return True, "Connected to %s on port %s" % (
                    settings.modeS_host,
                    settings.modeS_port,
                )
            else:
                return True, "Not connected"
        elif flag:
            self.connectToHost(settings.modeS_host, settings.modeS_port)
            stack.stack("OP")
            return True, "Connecting to %s on port %s" % (
                settings.modeS_host,
                settings.modeS_port,
            )
        else:
            self.disconnectFromHost()

        return True


class Decoder:
    import numpy as np

    # fmt: off
    MODES_CHECKSUM_TABLE = [
        0x3935ea, 0x1c9af5, 0xf1b77e, 0x78dbbf,
        0xc397db, 0x9e31e9, 0xb0e2f0, 0x587178,
        0x2c38bc, 0x161c5e, 0x0b0e2f, 0xfa7d13,
        0x82c48d, 0xbe9842, 0x5f4c21, 0xd05c14,
        0x682e0a, 0x341705, 0xe5f186, 0x72f8c3,
        0xc68665, 0x9cb936, 0x4e5c9b, 0xd8d449,
        0x939020, 0x49c810, 0x24e408, 0x127204,
        0x093902, 0x049c81, 0xfdb444, 0x7eda22,
        0x3f6d11, 0xe04c8c, 0x702646, 0x381323,
        0xe3f395, 0x8e03ce, 0x4701e7, 0xdc7af7,
        0x91c77f, 0xb719bb, 0xa476d9, 0xadc168,
        0x56e0b4, 0x2b705a, 0x15b82d, 0xf52612,
        0x7a9309, 0xc2b380, 0x6159c0, 0x30ace0,
        0x185670, 0x0c2b38, 0x06159c, 0x030ace,
        0x018567, 0xff38b7, 0x80665f, 0xbfc92b,
        0xa01e91, 0xaff54c, 0x57faa6, 0x2bfd53,
        0xea04ad, 0x8af852, 0x457c29, 0xdd4410,
        0x6ea208, 0x375104, 0x1ba882, 0x0dd441,
        0xf91024, 0x7c8812, 0x3e4409, 0xe0d800,
        0x706c00, 0x383600, 0x1c1b00, 0x0e0d80,
        0x0706c0, 0x038360, 0x01c1b0, 0x00e0d8,
        0x00706c, 0x003836, 0x001c1b, 0xfff409,
        0x000000, 0x000000, 0x000000, 0x000000,
        0x000000, 0x000000, 0x000000, 0x000000,
        0x000000, 0x000000, 0x000000, 0x000000,
        0x000000, 0x000000, 0x000000, 0x000000,
        0x000000, 0x000000, 0x000000, 0x000000,
        0x000000, 0x000000, 0x000000, 0x000000
    ]
    # fmt: on

    @staticmethod
    def hex2bin(hexstr):
        """Convert a hexadecimal string to binary string, with zero fillings."""
        length = len(hexstr) * 4
        msgbin = bin(int(hexstr, 16))[2:]
        return msgbin.zfill(length)

    @staticmethod
    def bin2int(binstr):
        return int(binstr, 2)

    @staticmethod
    def hex2int(hexstr):
        return int(hexstr, 16)

    @staticmethod
    def checksum(msg):
        if len(msg) == 28:
            offset = 0
        elif len(msg) == 14:
            offset = 112 - 56
        else:
            return False
        msgbin = Decoder.hex2bin(msg)
        chk = int(msg[22:28], 16)
        crc = 0
        for i in range(len(msgbin)):
            if int(msgbin[i]):
                crc ^= Decoder.MODES_CHECKSUM_TABLE[i + offset]
        return crc == chk

    @staticmethod
    def get_df(msg):
        msgbin = Decoder.hex2bin(msg)
        return Decoder.bin2int(msgbin[0:5])

    @staticmethod
    def get_ca(msg):
        msgbin = Decoder.hex2bin(msg)
        return Decoder.bin2int(msgbin[5:8])

    @staticmethod
    def get_icao_addr(msg):
        return msg[2:8]

    @staticmethod
    def get_tc(msg):
        msgbin = Decoder.hex2bin(msg)
        return Decoder.bin2int(msgbin[32:37])

    @staticmethod
    def get_oe_flag(msg):
        msgbin = Decoder.hex2bin(msg)
        return msgbin[53]

    @staticmethod
    def get_alt(msg):
        msgbin = Decoder.hex2bin(msg)
        q = msgbin[47]
        if q == "1":
            n = Decoder.bin2int(msgbin[40:47] + msgbin[48:52])
            return n * 25 - 1000
        else:
            return None

    @staticmethod
    def get_cprlat(msg):
        msgbin = Decoder.hex2bin(msg)
        return Decoder.bin2int(msgbin[54:71])

    @staticmethod
    def get_cprlon(msg):
        msgbin = Decoder.hex2bin(msg)
        return Decoder.bin2int(msgbin[71:88])

    @staticmethod
    def get_speed_heading(msg):
        msgbin = Decoder.hex2bin(msg)
        v_ew_dir = Decoder.bin2int(msgbin[45])
        v_ew = Decoder.bin2int(msgbin[46:56])
        v_ns_dir = Decoder.bin2int(msgbin[56])
        v_ns = Decoder.bin2int(msgbin[57:67])
        v_ew = -v_ew if v_ew_dir else v_ew
        v_ns = -v_ns if v_ns_dir else v_ns
        speed = Decoder.np.sqrt(v_ns * v_ns + v_ew * v_ew)
        heading = Decoder.np.arctan2(v_ew, v_ns)
        heading = heading * 360.0 / (2 * Decoder.np.pi)
        heading = heading if heading >= 0 else heading + 360
        return [speed, heading]

    @staticmethod
    def get_callsign(msg):
        chars = "#ABCDEFGHIJKLMNOPQRSTUVWXYZ#####_###############0123456789######"
        msgbin = Decoder.hex2bin(msg)
        csbin = msgbin[40:96]
        cs = ""
        for i in range(0, 48, 6):
            cs += chars[Decoder.bin2int(csbin[i : i + 6])]
        return cs.replace("_", "").replace("#", "")

    @staticmethod
    def get_position(msg0, msg1, t0, t1):
        cprlat0 = Decoder.get_cprlat(msg0)
        cprlat1 = Decoder.get_cprlat(msg1)
        cprlon0 = Decoder.get_cprlon(msg0)
        cprlon1 = Decoder.get_cprlon(msg1)
        return Decoder.cpr2position(cprlat0, cprlat1, cprlon0, cprlon1, t0, t1)

    @staticmethod
    def cpr2position(cprlat0, cprlat1, cprlon0, cprlon1, t0, t1):
        cprlat_even = cprlat0 / 131072.0
        cprlat_odd = cprlat1 / 131072.0
        cprlon_even = cprlon0 / 131072.0
        cprlon_odd = cprlon1 / 131072.0
        air_d_lat_even = 360.0 / 60
        air_d_lat_odd = 360.0 / 59
        j = int(59 * cprlat_even - 60 * cprlat_odd + 0.5)
        lat_even = air_d_lat_even * ((j % 60) + cprlat_even)
        lat_odd = air_d_lat_odd * ((j % 59) + cprlat_odd)
        if lat_even >= 270:
            lat_even -= 360
        if lat_odd >= 270:
            lat_odd -= 360
        if Decoder.cprNL(lat_even) != Decoder.cprNL(lat_odd):
            return None
        if t0 > t1:
            ni = Decoder.cprN(lat_even, 0)
            m = Decoder.np.floor(
                cprlon_even * (Decoder.cprNL(lat_even) - 1)
                - cprlon_odd * Decoder.cprNL(lat_even)
                + 0.5
            )
            lon = (360.0 / ni) * ((m % ni) + cprlon_even)
            lat = lat_even
        else:
            ni = Decoder.cprN(lat_odd, 1)
            m = Decoder.np.floor(
                cprlon_even * (Decoder.cprNL(lat_odd) - 1)
                - cprlon_odd * Decoder.cprNL(lat_odd)
                + 0.5
            )
            lon = (360.0 / ni) * ((m % ni) + cprlon_odd)
            lat = lat_odd
        if lon > 180:
            lon -= 360
        return [lat, lon]

    @staticmethod
    def cprN(lat, is_odd):
        nl_val = Decoder.cprNL(lat) - is_odd
        return nl_val if nl_val > 1 else 1

    @staticmethod
    def cprNL(lat):
        try:
            nz = 60
            a = 1 - Decoder.np.cos(Decoder.np.pi * 2 / nz)
            b = Decoder.np.cos(Decoder.np.pi / 180.0 * abs(lat)) ** 2
            nl_val = 2 * Decoder.np.pi / (Decoder.np.acos(1 - a / b))
            return int(nl_val)
        except Exception:
            return 1



""" BlueSky implementation of signals that can trigger one or more functions
    when a signal is emitted. """
from bluesky.core.funcobject import FuncObject


class SignalFactory(type):
    ''' Factory meta-class for Signal objects in BlueSky. '''
    # Store all signal objects in process
    __signals__ = dict()

    def __call__(cls, topic='', *args, **kwargs):
        ''' Factory function for Signal construction. '''
        # if no name is passed, return an anonymous Signal
        if not topic:
            return super().__call__('ANONYMOUS', *args, **kwargs)
        # Convert name to string if necessary
        if isinstance(topic, bytes):
            topic = topic.decode()
        # Signal topics are case-insensitive
        topic = topic.upper()
        # Check if this Signal already exists. If it doesn't, create it.
        sig = SignalFactory.__signals__.get(topic)
        if sig is None:
            sig = super().__call__(topic, *args, **kwargs)
            cls.__signals__[topic] = sig
        return sig


class Signal(metaclass=SignalFactory):
    """ A signal can trigger one or more functions when it is emitted. """
    def __init__(self, topic=''):
        self.topic = topic
        self.subscribers = list()

    def emit(self, *args, **kwargs):
        """ Trigger the registered functions with passed arguments. """
        # Iterate over shallow copy of subscriber list to avoid issues with connects/disconnects
        # that are called within one of our subscribed callback functions
        for sub in list(self.subscribers):
            sub(*args, **kwargs)

    def connect(self, func):
        """ Connect a new function to this signal. """
        self.subscribers.append(FuncObject(func))

    def disconnect(self, func):
        """ Disconnect a function from this signal. """
        try:
            self.subscribers.remove(FuncObject(func))
        except AttributeError:
            print(f'Function {func.__name__} is not connected to a signal.')
        except ValueError:
            print(f'Function {func.__name__} is not connected to Signal {self.topic}.')


def subscriber(func=None, topic=''):
    ''' BlueSky Signal subscription decorator.

        Functions decorated with this decorator will be called whenever the
        corresponding Signal emits data.

        Arguments:
        - topic: The topic to subscribe to for this function
    '''
    def deco(func):
        # Subscribe to topic.
        Signal(topic or func.__name__.upper()).connect(func)

        # Construct the subscription object, but return the original function
        return func
    # Allow both @subscriber and @subscriber(args)
    return deco if func is None else deco(func)



''' BlueSky resource access '''
import shutil
import itertools
from pathlib import Path
from importlib import import_module

try:
    from importlib.resources import files
    from importlib.readers import MultiplexedPath
except ImportError:
    # Python < 3.9 only provides deprecated resources API
    # Previous alternatives to find resources seem to be difficult
    # to get right on all platforms and python versions, so instead
    # this approach based on the search locations from the package spec
    from importlib.util import find_spec
    def files(package):
        res = ''
        if '.' in package:
            package, res = package.split('.', 1)
        s = find_spec(package)
        if s.submodule_search_locations:
            p = Path(s.submodule_search_locations[0])
        else:
            p = Path(package)
        return p / res.replace('.', '/')


    class MultiplexedPath:
        def __init__(self, *paths) -> None:
            self._paths = list(map(Path, paths))
        def iterdir(self):
            visited = []
            for path in self._paths:
                for file in path.iterdir():
                    if file.name in visited:
                        continue
                    visited.append(file.name)
                    yield file



class ResourcePath(MultiplexedPath):
    def __init__(self, *paths):
        base = files('bluesky.resources')
        paths = list(paths) + (base._paths if isinstance(base, MultiplexedPath) else [base])
        super().__init__(*paths)

    def appendpath(self, path):
        self._paths.append(path)

    def insertpath(self, path, pos=0):
        self._paths.insert(pos, path)

    def bases(self):
        for p in self._paths:
            yield p

    def base(self, idx):
        return self._paths[idx]

    @property
    def nbases(self):
        return len(self._paths)

    def as_posix(self, idx=0):
        return self._paths[idx].as_posix()

    def glob(self, pattern: str):
        files = set()
        for res in itertools.chain.from_iterable(p.glob(pattern) for p in self._paths):
            if res.is_file() and res.name in files:
                continue
            files.add(res.name)
            yield res

    def joinpath(self, *descendants):
        # first try to find child in current paths
        paths = []
        if not descendants:
            return self
        
        for path in (p.joinpath(*descendants) for p in self._paths):
            if path.exists():
                if path.is_dir():
                    # If it's a dir, try to combine it with others
                    paths.append(path)
                else:
                    # If it's a file, immediately return it
                    return path

        # if it does not exist, construct it with the first path
        return ResourcePath(*paths) if len(paths) > 1 else \
            paths[0] if len(paths) == 1 else self._paths[0].joinpath(*descendants)
    
    __truediv__ = joinpath


def resource(*descendants):
    ''' Get a path pointing to a BlueSky resource.
    
        Arguments:
        - descendants: Zero or more path-like objects (Path or str)

        Returns:
        - Path pointing to resource (file or directory)
          If arguments form an absolute path it is returned directly,
          otherwise a path relative to BlueSky's resource paths is returned.
    '''
    ret = Path(*descendants)
    if ret.is_absolute():
        return ret

    return resource.path.joinpath(*descendants)
resource.path = ResourcePath()


def init(workdir=None):
    ''' Initialise BlueSky search paths for resources and plugins. '''
    if workdir is None:
        if files('bluesky').parent == Path.cwd():
            # Assume BlueSky is running from source, e.g., cloned from GitHub
            # In this case additional resources (scenarios, plugins)
            # are found in the current working directory
            workdir = Path.cwd()
        else:
            # Assume BlueSky is running as a pip package
            # In this case additional resources are found in $home/bluesky
            workdir = Path.home().joinpath('bluesky')

            # Initialise this folder if it doesn't exist
            if not workdir.exists():
                # Create directory and populate with basics
                print(f'Creating BlueSky base directory "{workdir.absolute()}"')
                workdir.mkdir()

    elif not Path(workdir).exists():
        print(f"Specified working directory {workdir} doesn't exist!")

    # Ensure existence of scenario, plugins, output, and cache directories
    for subdir in map(workdir.joinpath, ('scenario', 'plugins', 'output', 'cache')):
        if not subdir.exists():
            print(f'Creating directory "{subdir}"')
            subdir.mkdir()

    # Ensure existence of config file
    cfgfile = workdir.joinpath("settings.cfg")
    if not cfgfile.exists():
        print(f'Copying default configfile to {cfgfile}')
        shutil.copy(resource('default.cfg'), cfgfile)

    # Set correct search paths for plugins
    plugins = import_module('bluesky.plugins')
    plugins.__spec__.submodule_search_locations.insert(0, workdir.joinpath('plugins').as_posix())

    # Set correct search paths for resource function
    resource.path = ResourcePath(workdir)



''' Stack Command implementation. '''
import inspect
from typing import Dict

from bluesky.core.funcobject import FuncObject
from bluesky.network.publisher import state_publisher
from bluesky.stack.argparser import Parameter, getnextarg, ArgumentError


class Command:
    ''' Stack command object. '''
    # Dictionary with all command objects
    cmddict: Dict[str, 'Command'] = dict()

    @staticmethod
    @state_publisher(topic='STACKCMDS')
    def pubcmdlist():
        ''' Send a dictionary with available stack commands when requested. '''
        return {'cmddict': {cmd : val.brief[len(cmd) + 1:] for cmd, val in Command.cmddict.items()}}

    @classmethod
    def addcommand(cls, func, parent=None, name='', **kwargs):
        ''' Add 'func' as a stack command. '''
        # Stack command name
        fname = inspect.unwrap(func).__name__.upper()
        name = (name.upper() or fname)

        # When a parent is passed this function is a subcommand or alt command
        target = Command.cmddict if parent is None else parent

        # Check if this command already exists
        cmdobj = target.get(name)
        if not cmdobj:
            # If command doesn't exist yet create it, and put it in the command
            # dict under its name and all its aliases
            cmdobj = cls(func, name, **kwargs)
            target[name] = cmdobj
            for alias in cmdobj.aliases:
                target[alias] = cmdobj
        elif cls is CommandGroup and isinstance(cmdobj, CommandGroup):
            # Multiple calls to @commandgroup should add command alternatives
            # TODO: @commandgroup repeated calls in subclasses should not create
            # alt commands
            altcmd = Command(func, name, parent=cmdobj, **kwargs)
            cmdobj.altcmds[func.__name__.upper()] = altcmd
            cmdobj = altcmd
        else:
            # for subclasses reimplementing stack functions we keep only one
            # Command object
            print(f'Attempt to reimplement {name} from {cmdobj.callback} to {func}')
            if not isinstance(cmdobj, cls):
                raise TypeError(f'Error reimplementing {name}: '
                                f'A {type(cmdobj).__name__} cannot be '
                                f'reimplemented as a {cls.__name__}')

    def __init__(self, func, name='', parent=None, **kwargs):
        self.name = name
        self.help = inspect.cleandoc(kwargs.get('help', ''))
        self.brief = kwargs.get('brief', '')
        self.aliases = kwargs.get('aliases', tuple())
        self.annotations = get_annot(kwargs.get('annotations', ''))
        self.params = list()
        self._callback = None
        self.callback = func

    def __call__(self, argstring):
        args = []
        param = None
        # Use callback-specified parameter parsers to generate param list from strings
        for param in self.params:
            result = param(argstring)
            argstring = result[-1]
            args.extend(result[:-1])

        # Parse repeating final args
        while argstring:
            if param is None or not param.gobble:
                msg = f'{self.name} takes {len(self.params)} argument'
                if len(self.params) > 1:
                    msg += 's'
                count = len(self.params)
                while argstring:
                    _, argstring = getnextarg(argstring)
                    count += 1
                msg += f', but {count} were given'
                raise ArgumentError(msg)
            result = param(argstring)
            argstring = result[-1]
            args.extend(result[:-1])

        # Call callback function with parsed parameters
        ret = self.callback(*args)
        # Always return a tuple with a success value and a message string
        if ret is None:
            return True, ''
        if isinstance(ret, (tuple, list)) and ret:
            if len(ret) > 1:
                # Assume that (success, echotext) is returned
                return ret[:2]
            ret = ret[0]
        return ret, ''

    def __repr__(self):
        if self.valid:
            return f'<Stack Command {self.name}, callback={self.callback}>'
        return f'<Stack Command {self.name} (invalid), callback=unbound method {self.callback}'

    @property
    def callback(self):
        ''' Callback pointing to the actual function that implements this
            stack command.
        '''
        return self._callback

    @callback.setter
    def callback(self, function):
        self._callback = FuncObject(function)
        spec = inspect.signature(function)
        self.brief = self.brief or (
            self.name + ' ' + ','.join(spec.parameters))
        self.help = self.help or inspect.cleandoc(
            inspect.getdoc(function) or '')
        paramspecs = list(filter(Parameter.canwrap, spec.parameters.values()))
        if self.annotations:
            self.params = list()
            pos = 0
            for annot, isopt in self.annotations:
                if annot == '...':
                    if paramspecs[-1].kind != paramspecs[-1].VAR_POSITIONAL:
                        raise IndexError('Repeating arguments (...) given for function'
                                            ' not ending in starred (variable-length) argument')
                    self.params[-1].gobble = True
                    break

                param = Parameter(paramspecs[pos], annot, isopt)
                if param:
                    pos = min(pos + param.size(), len(paramspecs) - 1)
                    self.params.append(param)
            if len(self.params) > len(paramspecs) and \
                paramspecs[-1].kind != paramspecs[-1].VAR_POSITIONAL:
                raise IndexError(f'More annotations given than function '
                                    f'{self.callback.__name__} has arguments.')
        else:
            self.params = [p for p in map(Parameter, paramspecs) if p]

    def helptext(self, subcmd=''):
        ''' Return complete help text. '''
        msg = f'<div style="white-space: pre;">{self.help}</div>\nUsage:\n{self.brief}'
        if self.aliases:
            msg += ('\nCommand aliases: ' + ','.join(self.aliases))
        return msg + '\n' + self.callback.info()

    def brieftext(self):
        ''' Return the brief usage text. '''
        return self.brief


class CommandGroup(Command):
    ''' Stack command group object.
        Command groups can have subcommands.
    '''
    def __init__(self, func, name='', **kwargs):
        super().__init__(func, name, **kwargs)
        self.subcmds = dict()
        self.altcmds = dict()

        # Give the function a method to add subcommands
        func.subcommand = lambda fun=None, **kwargs: command(
            fun, parent=self.subcmds, **kwargs)

        # Give the function a method to add alternative command implementations
        func.altcommand = lambda fun=None, **kwargs: command(
            fun, parent=self.altcmds, **kwargs)

    def __call__(self, strargs):
        # First check subcommand
        if strargs:
            subcmd, subargs = getnextarg(strargs)
            subcmdobj = self.subcmds.get(subcmd.upper())
            if subcmdobj:
                return subcmdobj(subargs)

        ret = super().__call__(strargs)
        msg = ret[1] if isinstance(ret, tuple) else ''
        success = ret[0] if isinstance(ret, tuple) else \
                  ret if isinstance(ret, bool) else True

        if not success:
            for altcmdobj in self.altcmds.values():
                ret = altcmdobj(strargs)
                success = ret[0] if isinstance(ret, tuple) else \
                          ret if isinstance(ret, bool) else True
                if isinstance(ret, tuple):
                    msg += '\n' + ret[1]
                if success:
                    return ret
        return success, msg


    def helptext(self, subcmd=''):
        ''' Return complete help text. '''
        if subcmd:
            obj = self.subcmds.get(subcmd)
            return obj.helptext() if obj else f'{subcmd} is not a subcommand of {self.name}'

        msg = super().helptext()
        if self.subcmds:
            msg += '\nSubcommands:'
            for cmdobj in self.subcmds.values():
                msg += f'\n{self.name} {cmdobj.brief} ({cmdobj.callback.info()})'
        if self.altcmds:
            msg += '\nAlternative command implementations:'
            for cmdobj in self.altcmds.values():
                msg += f'\n{self.name} {cmdobj.brief} ({cmdobj.callback.info()})'

        return msg

    def brieftext(self):
        ''' Return the brief usage text. '''
        msg = self.brief
        for subcmd in self.subcmds.values():
            msg += f'\n{self.name} {subcmd.brief}'
        return msg

def commandgroup(fun=None, parent=None, **kwargs):
    ''' BlueSky stack command decorator for command groups.

        Functions decorated with this decorator become available as stack
        functions, and have the ability to have subcommands.
    '''
    return command(fun, CommandGroup, parent, **kwargs)


def command(fun=None, cls=Command, parent=None, **kwargs):
    ''' BlueSky stack command decorator.

        Functions decorated with this decorator become available as stack
        functions.
    '''
    def deco(fun):
        # Construct the stack command object, but return the original function
        cls.addcommand(fun, parent, **kwargs)
        return fun
    # Allow both @command and @command(args)
    return deco(fun) if fun else deco


def append_commands(newcommands, syndict=None):
    """ Append additional functions to the stack command dictionary """
    for name, (brief, annotations, fun, hlp) in newcommands.items():
        if syndict:
            aliases = tuple(k for k,v in syndict.items() if v == name)
        else:
            aliases = tuple()
        # Use the command decorator function to register each new command
        Command.addcommand(fun, name=name, annotations=annotations, brief=brief, help=hlp, aliases=aliases)


def remove_commands(commands):
    """ Remove functions from the stack """
    for cmd in commands:
        Command.cmddict.pop(cmd)


def get_commands():
    """ Return the stack dictionary of commands. """
    return Command.cmddict


def get_annot(annotations):
    ''' Get annotations from string, or tuple/list. '''
    if isinstance(annotations, (tuple, list)):
        return tuple(annotations)
    # Assume it is a comma-separated string
    argtypes = []

    # Process and reduce annotation string from left to right
    # First cut at square brackets, then take separate argument types
    while annotations:
        opt = annotations[0] == "["
        cut = (annotations.find("]") if opt else annotations.find(
            "[") if "[" in annotations else len(annotations))

        types = annotations[:cut].strip("[,]").split(",")
        # Returned argtypes are tuples of type and optional status
        argtypes += zip(types, [opt or t == "..." for t in types])
        annotations = annotations[cut:].lstrip(",]")

    return tuple(argtypes)
